// file : GameState.cpp

#include <olcPixelGameEngine.h> 
#include "GameState.h"
#include "States/intro.h"
#include "States/game.h"
#include "States/levelTransition.h"
#include "States/loose.h"
#include "States/menu.h"
#include <olcNoiseMaker.h> // let's do that synth thing

#include "Sound.h"         // software synthisizer module
Synth::Sequencer *sequencer = nullptr;

//#define _RELEASE                                                   // to activate introState
std::shared_ptr<Core::States::Intro>          introState;
std::shared_ptr<Core::States::Menu>           menuState;
std::shared_ptr<Core::States::LevelTranstion> levelState;
std::shared_ptr<Core::States::LevelLoose>     looseState;
std::shared_ptr<Core::States::Game>           gameState;

auto clock_old_time = chrono::high_resolution_clock::now();
auto clock_real_time = chrono::high_resolution_clock::now();
double wallTime = 0.0;

void Core::StateManager::initialize(olc::PixelGameEngine* engine, vec2 clientSize)
{
	this->engine = engine; // core
	running = true;

	// sound synth startup  -------------------------------------------------
	std::vector<std::string> soundDevices = olcNoiseMaker<short>::Enumerate();
	//for (auto d : soundDevices) cout << std::string("Found output device : ") << d << std::endl;
	soundEngine = std::make_unique<olcNoiseMaker<short>>(soundDevices[0], 44100, 1, 8, 256);
	//Sound::dVolume = this->dVolume;
	soundEngine->SetUserFunction(Synth::MakeNoise);

	// drums ...pluck yeah!!!
	sequencer = new Synth::Sequencer(118.25, 4, 4, 3);
	sequencer->AddInstrument(&Synth::inst_kick);
	sequencer->AddInstrument(&Synth::inst_snare);
	sequencer->AddInstrument(&Synth::inst_hihat);
	sequencer->AddInstrument(&Synth::inst_cymb);
	sequencer->AddInstrument(&Synth::inst_bell);
	sequencer->AddInstrument(&Synth::inst_harmonica);
	sequencer->AddInstrument(&Synth::inst_harmonica2);
	sequencer->AddInstrument(&Synth::inst_harmonica3);

	// Author : mark kughler (goliath forge)
	// Lick : Do Diddley      |_............_||_............_||_............_||_............_|
	std::string strKick    = "....X...........X...X.........X.X...X...X...X...";
	std::string strSnare   = "......X...............X.........................";
	std::string strHiHat   = "..X.XX....X...X...X.......X...X...X...X...X...X.";
	std::string strCymb    = ".........................e....................X.";
	std::string strBell    = "......c...e...g.................................";
	std::string strHarmon3 = "....ag........c.......b.....a.........b.......e.";
	std::string strHarmon2 = "..fd......e...e..b..d.e.......a.c.......g...c.c.";
	std::string strHarmon  = ".b......................f.g.a.....g.a.b.........";
	// lower case notes (c major) - upper case sharp
	// c = c d e f g a b c
	// d = d e F g a b C d
	// e = e F G a b C D e
	// f = f g a A c d e f
	// g = g a b c d e F g
	// a = a b C d e F G a
	// b = b C D e F G A b

	//https://www.desmos.com/calculator
	sequencer->vecChannel.at(0).strBeat = strKick;
	sequencer->vecChannel.at(1).strBeat = strSnare;
	sequencer->vecChannel.at(2).strBeat = strHiHat;
	sequencer->vecChannel.at(3).strBeat = strCymb;
	sequencer->vecChannel.at(4).strBeat = strBell;
	sequencer->vecChannel.at(5).strBeat = strHarmon;
	sequencer->vecChannel.at(6).strBeat = strHarmon2;
	sequencer->vecChannel.at(7).strBeat = strHarmon3;

	// managed game screens ---------------------------------------------------
	introState = std::make_shared<States::Intro>(this, clientSize);
	menuState = std::make_shared<States::Menu>(this, clientSize);
	levelState = std::make_shared<States::LevelTranstion>(this, clientSize);
	looseState = std::make_shared<States::LevelLoose>(this, clientSize);
	gameState = std::make_shared<States::Game>(this, clientSize);

	// sprite assets -----------------------------------------------------------
	auto onImageLoad = [](std::string path, std::vector<olc::Sprite*> &spriteList) 
	{   // image loader
		olc::Sprite* sprite = new olc::Sprite();
		sprite->LoadFromFile(path);
		spriteList.push_back(sprite);
	};
	std::vector<std::string> imageNames = { // "Assets/SpinningOrb/"
		"Red/frame1", "Red/frame2", "Red/frame3", "Red/frame4", "Red/frame5", "Red/frame6", 
		"Green/frame1", "Green/frame2", "Green/frame3", "Green/frame4", "Green/frame5", "Green/frame6",
		"Blue/frame1", "Blue/frame2", "Blue/frame3", "Blue/frame4", "Blue/frame5", "Blue/frame6",
		"Yellow/frame1", "Yellow/frame2", "Yellow/frame3", "Yellow/frame4", "Yellow/frame5", "Yellow/frame6"
	};
	std::vector<std::string> animNames = { // "Assets/Actors/
		"soldier_block01", "soldier_block02", "soldier_block03", "soldier_block04",
		"soldier_look01", "soldier_look02", "soldier_look03", "soldier_look04", "soldier_look05", "soldier_look06", "soldier_look07", "soldier_look08", "soldier_look09",
	}; 

	for (auto name : imageNames) { onImageLoad("Assets/SpinningOrb/" + name + ".png", orbSprites); }
	for (auto name : animNames)  { onImageLoad("Assets/Actors/" + name + ".png", charSprites);     }

	


	states.push_back(menuState);
	// pre compile heirarchy addition
#ifdef _RELEASE
	states.push_back(introState);
#endif
	// --------------------------------------------------------------------------
}

void Core::StateManager::shutdown()
{
	introState->shutdown();
	menuState->shutdown();
	levelState->shutdown();
	looseState->shutdown();
	gameState->shutdown();
	
	soundEngine->Stop();
	if (sequencer != nullptr) { delete sequencer; sequencer = nullptr; }
}

void Core::StateManager::changeState(Core::States::Base_state* _state) { }
void Core::StateManager::pushState(Core::States::Base_state* _state)   { }
void Core::StateManager::popState() { states.pop_back();     }
void Core::StateManager::draw()     { states.back()->draw(); }

// event loop controller 
void Core::StateManager::handleEvents(eStateAction _action)
{
	switch (_action) {
	case eStateAction::pause:
		if (states.back()->name == "game_state")
			states.pop_back();
		break;
	case eStateAction::play:
		if (states.back()->name == "menu_state")
		{	int w = engine->GetDrawTargetWidth();
			int h = engine->GetDrawTargetHeight();
			
			sequencer->isPlaying = false;
			RECT rect = { 0, 0, w, h };
			gameState->reset(0);
			looseState->reset();
			states.push_back(gameState);
		}
		if (states.back()->name == "level_transition_state")
			states.pop_back();
		break;
	case eStateAction::levelUp:
		gameState->reset(levelState->level); // advance game 
		states.push_back(levelState);        // resume level up
		break;
	case eStateAction::win:
		if (states.back()->name == "game_state")
		{	MessageBox(0, "you win", "", 0);
			//winState->winnerName = gameState->getWinnerName();
			//winState->winningTime = gameState->getTimeString();
			//states.push_back(winState);
		}
		break;
	case eStateAction::loose:
		if (states.back()->name == "game_state")
		{	// todo : make this deferred execution (or fade out somehow)
			levelState->level = 0;
			sequencer->isPlaying = true;
			states.pop_back();             // remove game
			states.push_back(looseState);  // resume loose
		}
		break;
	case eStateAction::quit:
		if (states.back()->name == "menu_state")
			PostQuitMessage(0);

		if (states.back()->name == "level_loose_state")
			states.pop_back();

		if (states.back()->name == "game_state")
		{	::ShowCursor(true);
			::ClipCursor(nullptr);
			states.pop_back();
		}
		/*
		if (states.back()->name == "score_state")
			states.pop_back();

		if (states.back()->name == "win_state")
		{	// todo : record score
			gameState->reset(0);
			states.pop_back();
		}*/
		break;
	} // end switch(action)
}

void Core::StateManager::playSound(int k, int type, double offset_t, double duration_t)
{
	// Key has been pressed so create a new note
	Synth::Note n;
	n.on_t = soundEngine->GetTime() + offset_t;
	n.off_t = n.on_t + duration_t;
	n.isActive = true;
	k += 64;
	switch (type) {
	case explosion:
		n.channel = &Synth::inst_explosion;
		n.id = k;
		break;
	case bell:
		n.channel = &Synth::inst_bell;
		n.id = k;
		break;
	case tick:
		n.channel = &Synth::inst_hihat;
		n.id = k;
		break;
	}

	if (type < numSoundTypes && type > -1)
	{   // Add note to sound engine vector queue
		Synth::muxNotes.lock();
		auto noteFound = find_if(Synth::vecNotes.begin(), Synth::vecNotes.end(), [&k, &n](Synth::Note const& item) 
		{ return item.id == k && item.channel == n.channel; });

		if (noteFound == Synth::vecNotes.end())
		{ // end of list
			Synth::vecNotes.emplace_back(n);
		}
		Synth::muxNotes.unlock();
	}
}	

void Core::StateManager::update(float deltaTime)
{
	// screen ---------------
	//std::shared_ptr<States::Base_state> state = states.back();
	if (states.back()->name == "game_state")
	{
		float scale = 0.35f + (1.5f * this->difficulty);
		states.back()->update(deltaTime * scale);
	} else
		states.back()->update(deltaTime);
	
	// sound engine -----------------------------------------------
	Synth::volume_main = this->dVolume;
	clock_real_time = chrono::high_resolution_clock::now();
	auto clock_delta_time = clock_real_time - clock_old_time;
	clock_old_time = clock_real_time;

	double dt = chrono::duration<double>(clock_delta_time).count();
	wallTime += dt;
	double time_now = soundEngine->GetTime();

	int newNotes = sequencer->Update(dt);
	Synth::muxNotes.lock();
	for (int a = 0; a < newNotes; a++)
	{	sequencer->vecNotes[a].on_t = time_now;
		if (sequencer->vecNotes[a].id != 64)
			sequencer->vecNotes[a].off_t = time_now+0.0001; // toss slightly into the future

		Synth::vecNotes.emplace_back(sequencer->vecNotes[a]);
	}
	Synth::muxNotes.unlock();

	 // keyboard sound interface
	/*for (int k = 0; k < 16; k++)
	{
		short nKeyState = GetAsyncKeyState((unsigned char)("ZSXCFVGBNJMK\xbcL\xbe\xbf"[k]));

		// Check if note already exists in currently playing notes
		Synth::muxNotes.lock();
		auto noteFound = find_if(Synth::vecNotes.begin(), Synth::vecNotes.end(), [&k](Synth::Note const& item) { return item.id == k + 64 && item.channel == &Synth::inst_harmonica; });
		if (noteFound == Synth::vecNotes.end())
		{
			// Note not found in vector
			if (nKeyState & 0x8000)
			{
				// Key has been pressed so create a new note
				Synth::Note n;
				n.id = k + 64;
				n.on_t = time_now;
				n.isActive = true;
				n.channel = &Synth::inst_harmonica;

				// Add note to vector
				Synth::vecNotes.emplace_back(n);
			}
		}
		else
		{
			// Note exists in vector
			if (nKeyState & 0x8000)
			{
				// Key is still held, so do nothing
				if (noteFound->off_t > noteFound->on_t)
				{
					// Key has been pressed again during release phase
					noteFound->on_t = time_now;
					noteFound->isActive = true;
				}
			}
			else
			{
				// Key has been released, so switch off
				if (noteFound->off_t < noteFound->on_t)
					noteFound->off_t = time_now;
			}
		}
		Synth::muxNotes.unlock();
	}*/
}

