#ifndef PHYSICS_OBJECT_H_
#define PHYSICS_OBJECT_H_

#include <algorithm>

class PhysicsObject //-----------------------------------------------------------------
{                   //-------------------- abstract base interface --------------------
public:
	vec2 pos = { 0.f, 0.f };
	vec2 vel = { 0.f, 0.f };
	vec2 acc = { 0.f, 0.f };

	float mass = 1.f;
	float radius = 1.f;
	bool  bStable = false; // stopped moving? no. Sound trigger  (game.onUpdate)
	float friction = 0.8f;
	
	int nBounceCount = 0;
	int nBounceBeforDeath = -1; // -1 = infinate
	bool bDead = false;
	bool bExplode = false;
	size_t id;

	PhysicsObject(float x = 0.f, float y = 0.f)
	{
		pos = { x, y };
		bExplode = false;
	}

	virtual void Draw(olc::PixelGameEngine *engine, float fOffsetX, float fOffsetY) = 0;
	virtual int BounceDeathAction() = 0;

	vec2 nextPos;
	vec2 nextVel;
	bool bHitThisFrame = false;
	olc::Pixel color = olc::RED;
};




//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

#define dummyRadius 4.f

class BallGoal : public PhysicsObject
{
public: 
	BallGoal(float x = 0.f, float y = 0.f) : PhysicsObject(x, y)
	{
		mass = 10.f;
		radius = 8;
		color = olc::YELLOW;
	}

	virtual void Draw(olc::PixelGameEngine *engine, float offsetX, float offsetY)
	{
		vec2 p = { pos.x - offsetX, pos.y - offsetY };
		
		for (auto i : vecModel)
		{   // draw model pixels
			engine->Draw(int32_t(p.x + i.first), int32_t(p.y + i.second), color);
		}
		engine->Draw((int32_t)p.x, (int32_t)p.y, olc::RED);
	}

	virtual int BounceDeathAction() { return 0; }

	bool isCarried = true;
private:
	static std::vector<std::pair<float, float>> vecModel;
};



class BallBumper : public PhysicsObject
{
public:
	BallBumper(float x = 0.f, float y = 0.f) : PhysicsObject(x, y)
	{
		mass = 100.f;
		radius = 10;
		color = olc::DARK_GREY;
	}

	virtual void Draw(olc::PixelGameEngine *engine, float offsetX, float offsetY)
	{
		vec2 p = { pos.x - offsetX, pos.y - offsetY };
		for (auto i : vecModel)
		{   // draw model pixels
			engine->Draw(int32_t(p.x + i.first), int32_t(p.y + i.second), color);
		}
	}

	virtual int BounceDeathAction() { return 0; }

private:
	static std::vector<std::pair<float, float>> vecModel;
};

class BallFactory : public PhysicsObject
{
public:
	BallFactory(float x = 0.f, float y = 0.f) : PhysicsObject(x, y)
	{
		mass = 100.f;
		radius = 10;
		color = olc::Pixel(240, 240, 255, 255);
	}

	virtual void Draw(olc::PixelGameEngine *engine, float offsetX, float offsetY)
	{
		int px = int(pos.x - offsetX);
		int py = int(pos.y - offsetY);
	
		for (auto i : vecModel)
		{   // draw model pixels
			engine->Draw(int32_t(px + i.first), int32_t(py + i.second), color);
		}

		engine->SetPixelMode(olc::Pixel::ALPHA);
		// under lights
		int LL = -8; // x pos offset
		int LR = 5;
		for (int i = 0; i < 4; i++) {
			for (int j = 6; j < 8; j++) {
				engine->Draw(px + LL + i, py + j, light01);
				engine->Draw(px + LR + i, py + j, light02);
			}
		}
		engine->Draw(px - 7, py + 8, light01);
		engine->Draw(px - 6, py + 8, light01);// left
		engine->Draw(px + 6, py + 8, light02);
		engine->Draw(px + 7, py + 8, light02);// right

		for (int i = -2; i < 3; i++) {
			engine->Draw(px - i, py - 6, light);
			engine->Draw(px - i, py - 5, light);
		}
		engine->Draw(px - 1, py - 7, light);
		engine->Draw(px , py - 7, light);  // top flasher
		engine->Draw(px + 1, py - 7, light);

		engine->SetPixelMode(olc::Pixel::NORMAL);

		light.a -= 2; if (light.a < 0) light.a = 255;
		light01.a -= 5; if (light01.a < 0) light01.a = 255;
		light02.a -= 4; if (light02.a < 0) light02.a = 255;
	}

	virtual int BounceDeathAction() { return 0; }

	bool isMoving = true;
private:
	olc::Pixel light = {0,255, 0, 255 };
	olc::Pixel light01 = { 255, 0, 0, 255 };
	olc::Pixel light02 = { 255, 0, 0, 255 };
	static std::vector<std::pair<float, float>> vecModel;
};

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
// Debug marker
// These derived classes define appearance and behavior individually per class type
// The graphic used per type is static shared and factory pattern initialized.
// the update ---- wip todo yada

class Dummy : public PhysicsObject
{
public:
	Dummy(float x = 0.f, float y = 0.f) : PhysicsObject(x, y)
	{
		mass = 0.08f;
		radius = dummyRadius;
		nBounceBeforDeath = 8;
	}

	virtual void Draw(olc::PixelGameEngine *engine, float offsetX, float offsetY)
	{
		vec2 p = { vecModel[0].first - offsetX, vecModel[0].second - offsetY };
		int px, py;
		if (pos.x < 0.f) pos.x += 1600;
		if (pos.x > 1600.f) pos.x -= 1600;
		for( auto i : vecModel)
		{
			px = int(i.first - offsetX);
			py = int(i.second - offsetY);
			engine->DrawLine(int32_t(p.x + pos.x), int32_t(p.y + pos.y), int32_t(px + pos.x), int32_t(py + pos.y), color);
			p = { i.first - offsetX, i.second - offsetY };
		}

		p = { pos.x-offsetX, pos.y-offsetY };
		engine->DrawLine(int32_t(p.x), int32_t(p.y), int32_t(p.x+(nextVel.x*radius)), int32_t(p.y+(nextVel.y*radius)), olc::YELLOW);
	}

	virtual int BounceDeathAction() 
	{
		return 0; 
	}

private:
	static std::vector<std::pair<float, float>> vecModel;   
};
 



//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
// Debris : in game fragment marker 
// limited lifetime. (bounce count) 
//
class Debre : public PhysicsObject
{
public:
	Debre(float x = 0.f, float y = 0.f, float vx = 0.f, float vy = 0.f) : PhysicsObject(x, y) 
	{
		mass = 0.06f;
		radius = 0.005f;
		nextVel = { vx, vy };
		nBounceBeforDeath = 1;
		bDead = true;
		type = 0;
	}

	virtual void Draw(olc::PixelGameEngine * engine, float offsetX, float offsetY)
	{
		int px = int(pos.x - offsetX);
		int py = int(pos.y - offsetY);
	
		switch(type)
		{
		case 0:
			engine->Draw(px, py, olc::YELLOW);
			break;
		case 1:
			for (auto i : vecModel)
			{   // draw model pixels
				engine->Draw(int(px + i.first), int(py + i.second), color);
			}
			break;
		}
	}

	virtual int BounceDeathAction()
	{
		return 0;
	}

public:
	int type;
private:
	static std::vector<std::pair<float, float>> vecModel;
};

// ----------------------------------------------------------------------
//-----------------------------------------------------------------------
#endif
