#ifndef my_olc_sound_imp

//
// brief : https://www.youtube.com/watch?v=tgamhuQnOkM basic notes
//
//         https://www.youtube.com/watch?v=OSCzKOqtgcA oscillators and envelopes
//
//amplitude * sin(frequency) * time

std::unique_ptr<olcNoiseMaker<short>> soundEngine;

namespace Synth
{

	double volume_main;

	const double TwoPi = 2.0 * 3.14159;
	const double TwoOverPi = 2.0 / 3.14159;
	const double PiOverTwo = 3.14159 / 2.0;
	const double d12thRootOf2 = pow(2.0, 1.0 / 12.0); // <-- this m0fo (step generator)

	enum eWaveType { SineWave, SquareWave, TriangleWave, SawWave, SawWaveFast, Noise, numWaveTypes };
	double w(const double htz) { return htz * TwoPi; } // frequency to angular velocity
	double s(const int noteID) { return 8 * pow(1.0594630943592952645618252949463, noteID); }

#pragma region Note ------------------------------------------------------
	// Note Structure - abstract base class (virtual)
	struct Instrument;
	struct Note
	{
		int id; // position in scale
		double on_t; // start wall time
		double off_t; // stop wall time
		bool isActive;
		Instrument *channel;

		Note()
		{
			id = 0; 
			on_t = 0.0;
			off_t = 0.0;
			isActive = false;
			channel = nullptr;
		}
	};
	std::vector<Note> vecNotes;

#pragma endregion

#pragma region Oscillator ------------------------------------------------

	double osc(double dt, double htz, int type = SineWave, double lfoHtz = 0.0, double lfoAmp = 0.0, double gain = 50.0)
	{
		double vT = w(htz) * dt + lfoAmp * htz * (sin(w(lfoHtz) * dt));

		switch (type)
		{
		case SineWave: // (pops)
			return sin(vT);// * 1.6;

		case SquareWave:
			return sin(vT) > 0.0 ? 0.33 : -0.33; // balanced cap

		case TriangleWave: // (pops)
			return asin(sin(vT)) * TwoOverPi;

		case SawWave: // (expensive)
		{
			double d = 0.0; // accumulator
			for (double n = 1.0; n < gain; n++) // todo : rename gain 2019-04-29
				d += (sin(n * vT)) / n;
			return d * TwoOverPi;// * 0.7);
		}

		case SawWaveFast: // (faster)
			return TwoOverPi * (htz * 3.14159 * fmod(dt, 1.0 / htz) - PiOverTwo);// * 0.8); // balanced 0.8

		case Noise: //(-1)-->(+1) scaled
			return 2.0 * ((double)rand() / (double)RAND_MAX) - 1.0;

		default:
			return 0.0;
		}
	}

#pragma endregion
#pragma region Envelope --------------------------------------------------
	
	struct Envelope // abstract base
	{
		virtual double amplitude(const double dt, const double on_t, const double off_t) = 0;
	};

	struct EnvelopeADSR : public Envelope
	{
		double dAttackTime;
		double dDecayTime;
		double dReleaseTime;

		double dSustainAmplitude;
		double dStartAmplitude;

		EnvelopeADSR()
		{
			dAttackTime = 0.1;
			dDecayTime = 0.1;
			dReleaseTime = 0.2;
			dSustainAmplitude = 1.0;
			dStartAmplitude = 1.0;
		}

		virtual double amplitude(const double dt, const double on_t, const double off_t)
		{	// index dTime into the lifetime of the envelope
			double dAmplitude = 0.0;
			double releaseAmplitude = 0.0;

			if(on_t > off_t)
			{ 
				double lifeTime = dt - on_t;
				// attack
				if (lifeTime <= dAttackTime)
					dAmplitude = (lifeTime / dAttackTime) * dStartAmplitude; // normalized time of start
				// decay
				if (lifeTime > dAttackTime && lifeTime <= (dAttackTime + dDecayTime))
					dAmplitude = ((lifeTime - dAttackTime) / dDecayTime) 
					          * (dSustainAmplitude - dStartAmplitude) + dStartAmplitude; // decay amplitude gradient
				// sustain
				if (lifeTime > (dAttackTime + dDecayTime)) 
					dAmplitude = dSustainAmplitude; 
			}
			else
			{
				double lifeTime = off_t - on_t;

				if (lifeTime <= dAttackTime)
					releaseAmplitude = (lifeTime / dAttackTime) * dStartAmplitude;
				
				if (lifeTime > dAttackTime && lifeTime <= (dAttackTime + dDecayTime))
					releaseAmplitude = ((lifeTime - dAttackTime) / dDecayTime) * (dSustainAmplitude - dStartAmplitude) + dStartAmplitude;
				
				if (lifeTime > (dAttackTime + dDecayTime))
					releaseAmplitude = dSustainAmplitude;
				
				dAmplitude = ((dt - off_t) / dReleaseTime) * (0.0 - releaseAmplitude) + releaseAmplitude;
			}

			if (dAmplitude <= 0.01)
				dAmplitude = 0.0;

			return dAmplitude;
		}
	};

	double env(const double dt, Envelope &env, const double on_t, const double off_t)
	{
		return env.amplitude(dt, on_t, off_t);
	}

#pragma endregion
#pragma region Instruments -----------------------------------------------

	struct Instrument // abstract base
	{
		double dVolume;
		EnvelopeADSR env;
		double dMaxLifeTime;
		std::string name;

		virtual double sound(double dt, Synth::Note n, bool &finished) = 0;
	};

	struct Instrument_Bell : public Instrument
	{
		Instrument_Bell() 
		{
			env.dAttackTime = 0.01;
			env.dDecayTime = 1.0;
			env.dSustainAmplitude = 0.01;
			env.dReleaseTime = 1.0;
			dMaxLifeTime = 3.0;
			dVolume = 0.22;
			name = "Bell";
		}

		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (amplitude <= 0.0) finished = true;

			double sound = osc(dt - n.on_t, Synth::s(n.id + 12), SineWave, 5.0, 0.001)
				   + 0.5 * osc(dt - n.on_t, Synth::s(n.id + 24))
				  + 0.25 * osc(dt - n.on_t, Synth::s(n.id + 36));
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_Harmonica : public Instrument
	{
	public:
		Instrument_Harmonica()
		{
			env.dAttackTime = 0.0;
			env.dDecayTime = 1.0;
			env.dSustainAmplitude = 0.95;
			env.dReleaseTime = 0.1;
			dMaxLifeTime = -1.0;
			dVolume = 0.18;
			name = "Harmonica";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (amplitude <= 0.0) finished = true;

			double sound = osc(n.on_t - dt, Synth::s(n.id - 24), Synth::SawWave, 5.0, 0.001, 100)
				+ osc(dt - n.on_t, Synth::s(n.id - 12), Synth::SquareWave, 5.0, 0.001)
				+ 0.5 * osc(dt - n.on_t, Synth::s(n.id), Synth::SquareWave)
				+ 0.05 * osc(dt - n.on_t, Synth::s(n.id + 12), Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_Harmonica2 : public Instrument
	{
	public:
		Instrument_Harmonica2() 
		{
			env.dAttackTime = 0.0;
			env.dDecayTime = 1.0;
			env.dSustainAmplitude = 0.95;
			env.dReleaseTime = 0.1;
			dMaxLifeTime = -1.0;
			dVolume = 0.18;
			name = "Harmonica2";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{ 
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (amplitude <= 0.0) finished = true;
			
			double sound = osc(n.on_t - dt, Synth::s(n.id - 12), Synth::SawWave, 5.0, 0.001, 100)
				         + osc(dt - n.on_t, Synth::s(n.id), Synth::SquareWave, 5.0, 0.001)
				   + 0.5 * osc(dt - n.on_t, Synth::s(n.id + 12), Synth::SquareWave)
				  + 0.05 * osc(dt - n.on_t, Synth::s(n.id + 24), Synth::Noise);
			return amplitude * sound * dVolume; 
		}
	};
	struct Instrument_Harmonica3 : public Instrument
	{
	public:
		Instrument_Harmonica3()
		{
			env.dAttackTime = 0.0;
			env.dDecayTime = 1.0;
			env.dSustainAmplitude = 0.95;
			env.dReleaseTime = 0.1;
			dMaxLifeTime = -1.0;
			dVolume = 0.18;
			name = "Harmonic3";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (amplitude <= 0.0) finished = true;

			double sound = osc(n.on_t - dt, Synth::s(n.id), Synth::SawWave, 5.0, 0.001, 100)
				+ osc(dt - n.on_t, Synth::s(n.id + 12), Synth::SquareWave, 5.0, 0.001)
				+ 0.5 * osc(dt - n.on_t, Synth::s(n.id + 24), Synth::SquareWave)
				+ 0.05 * osc(dt - n.on_t, Synth::s(n.id + 36), Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_DrumKick : public Instrument
	{
		Instrument_DrumKick() 
		{
			env.dAttackTime = 0.01;
			env.dDecayTime = 0.15;
			env.dSustainAmplitude = 0.1;
			env.dReleaseTime = 0.001;
			dMaxLifeTime = 1.5;
			dVolume = 0.6;
			name = "Drum Kick";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.99 * osc(dt - n.on_t, Synth::s(n.id - 36), Synth::SineWave, 1.0, 1.0)
				         + 0.01 * osc(dt - n.on_t, 0, Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_DrumTom : public Instrument
	{
		Instrument_DrumTom()
		{
			env.dAttackTime = 0.001;
			env.dDecayTime = 0.15;
			env.dSustainAmplitude = 0.1;
			env.dReleaseTime = 0.001;
			dMaxLifeTime = 1.5;
			dVolume = 0.7;
			name = "Drum Tom";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.99 * osc(dt - n.on_t, Synth::s(n.id - 36), Synth::SineWave, 1.0, 1.0)
				+ 0.01 * osc(dt - n.on_t, 0, Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_DrumSnare : public Instrument
	{
		Instrument_DrumSnare() 
		{
			env.dAttackTime = 0.002;
			env.dDecayTime = 0.2;
			env.dSustainAmplitude = 0.0;
			env.dReleaseTime = 0.001;
			dMaxLifeTime = 1.0;
			dVolume = 1.0;
			name = "Drum Snare";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.5 * osc(dt - n.on_t, Synth::s(n.id - 24), Synth::SineWave, 0.5, 1.0)
				         + 0.5 * osc(dt - n.on_t, 0, Synth::Noise);
			return amplitude * sound * dVolume; 
		}
	};
	struct Instrument_DrumHiHat : public Instrument
	{
		Instrument_DrumHiHat() 
		{
			env.dAttackTime = 0.01;
			env.dDecayTime = 0.05;
			env.dSustainAmplitude = 0.1;
			env.dReleaseTime = 0.001;
			dMaxLifeTime = 0.12;
			dVolume = 0.3;
			name = "Drum HiHat";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{ 
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.2 * osc(dt - n.on_t, Synth::s(n.id - 12), Synth::SquareWave, 1.5, 1)
				         + 0.8 * osc(dt - n.on_t, 0, Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_DrumCymbal : public Instrument
	{
		Instrument_DrumCymbal()
		{
			env.dAttackTime = 0.01;
			env.dDecayTime = 0.05;
			env.dSustainAmplitude = 0.8;
			env.dReleaseTime = 0.1;
			dMaxLifeTime = 0.3;
			dVolume = 0.8;
			name = "Drum Cymbal";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.3 * osc(dt - n.on_t, Synth::s(n.id + 24), Synth::SquareWave, 1.5, 1)
				+ 0.7 * osc(dt - n.on_t, 0, Synth::Noise);
			return amplitude * sound * dVolume;
		}
	};
	struct Instrument_Explosion : public Instrument
	{
		Instrument_Explosion()
		{
			env.dAttackTime = 0.01;
			env.dDecayTime = 0.05;
			env.dSustainAmplitude = 0.8;
			env.dReleaseTime = 0.1;
			dMaxLifeTime = 0.3;
			dVolume = 0.6;
			name = "Explosion";
		}
		virtual double sound(const double dt, Synth::Note n, bool &finished)
		{
			double amplitude = Synth::env(dt, env, n.on_t, n.off_t);
			if (dMaxLifeTime > 0.0 && dt - n.on_t >= dMaxLifeTime)	finished = true;

			double sound = 0.75 * osc(dt - n.on_t, Synth::s(n.id - 14), Synth::SquareWave, 1.5, 1)
				+ 0.25 * osc(dt - n.on_t, 0, Synth::Noise);

			return amplitude * sound * dVolume;
		}
	};

	Synth::Instrument_Bell inst_bell;
	Synth::Instrument_Harmonica inst_harmonica;
	Synth::Instrument_Harmonica2 inst_harmonica2;
	Synth::Instrument_Harmonica3 inst_harmonica3;
	Synth::Instrument_DrumKick inst_kick;
	Synth::Instrument_DrumTom inst_tom;
	Synth::Instrument_DrumSnare inst_snare;
	Synth::Instrument_DrumHiHat inst_hihat;
	Synth::Instrument_DrumCymbal inst_cymb;
	Synth::Instrument_Explosion inst_explosion;
	// add more here....more...............
#pragma endregion
#pragma region Channel_Sequencer -----------------------------------------

	struct Sequencer
	{
	public:
		struct Channel
		{
			Instrument* instrument;
			std::string strBeat;
		};

	public:
		Sequencer(double tempo = 120.0, int beats = 4, int subbeats = 4, int measures=1)
		{
			nBeats = beats;
			nSubBeats = subbeats;
			dTempo = tempo;
			dBeatTime = (60.0f / tempo) / (float)subbeats;
			nCurrentBeat = 0;
			nTotalBeats = (nSubBeats * nBeats) * measures;
			dAccumulate = 0.0;
		}

		bool isPlaying = true;
		int Update(double elapsedTime)
		{
			if (isPlaying)
			{
				vecNotes.clear();
				dAccumulate += elapsedTime;
				while (dAccumulate >= dBeatTime)
				{
					dAccumulate -= dBeatTime;
					nCurrentBeat++;
					if (nCurrentBeat >= nTotalBeats)
						nCurrentBeat = 0; // wrap around

					int c = 0;
					for (auto v : vecChannel)
					{
						if (v.strBeat[nCurrentBeat] != '.')
						{
							int id;
							switch (v.strBeat[nCurrentBeat])
							{
							case 'X': id = 64; break;
							case 'B':
							case 'c': id = 67; break;
							case 'C': id = 68; break;
							case 'd': id = 69; break;
							case 'D': id = 70; break;
							case 'e': id = 71; break;
							case 'E':
							case 'f': id = 72; break;
							case 'F': id = 73; break;
							case 'g': id = 74; break;
							case 'G': id = 75; break;
							case 'a': id = 76; break;
							case 'A': id = 77; break;
							case 'b': id = 78; break;
							default: id = 0;
							}

							if (id > 0)
							{
								Note n;
								n.channel = vecChannel[c].instrument;
								n.isActive = true;
								n.id = id;
								vecNotes.push_back(n);
							}
						}
						c++;
					}
				}
				return (int)vecNotes.size();
			}
			return 0;
		}

		void AddInstrument(Instrument* inst)
		{
			Channel c;
			c.instrument = inst;
			vecChannel.push_back(c);
		}

	public:
		int nBeats;
		int nSubBeats;
		double dTempo;
		double dBeatTime;
		double dAccumulate;
		int nCurrentBeat;
		int nTotalBeats;

	public:
		std::vector<Channel> vecChannel;
		std::vector<Note> vecNotes;
	};

#pragma endregion

	mutex muxNotes; // sound thread lock flag

	typedef bool(*lambda)(Synth::Note const &item);
	template<class T>
	void safe_remove(T &v, lambda f)
	{
		auto n = v.begin();
		while (n != v.end())
			if (!f(*n))
				n = v.erase(n);
			else
				++n;
	}

	double MakeNoise(int nChannel, double dTime)
	{   //    note : real systems use integers
		// warning : this callback is called from a separate thread 
	    //    owner: GameState Manager

		unique_lock<mutex> lm(muxNotes);
		double output = 0.0;

		// mix the notes
		for (auto &n : vecNotes)
		{
			bool finished = false;
			double sound = 0.0;
			if (n.channel != nullptr)
				sound = n.channel->sound(dTime, n, finished);

			output += sound; // todo : maybe only add on not finished. else kill here.
			if (finished)
				n.isActive = false;
		}
		safe_remove<std::vector<Note>>(vecNotes, [](Note const &item) { return item.isActive; });
		return output * volume_main * 0.22; // scaled master volume speaker blow up
	}
}

#define my_olc_sound_imp
#endif
